{-# LANGUAGE DeriveDataTypeable #-}

module Examples.Haskell where

import Data.Generics
import Data.List (intersperse)

import CC.Syntax
import CC.Semantics
import CC.Static
import CC.Zipper
import Examples.Edit


-------------------------
-- Variational Haskell --
-------------------------

-- variational Haskell (vhaskell)
--
type VHaskell = V Haskell

-- expanded Haskell (xhaskell)
--
data Haskell = App Haskell Haskell
         | Var Name
         | Val Int
         | Fun Name [Haskell] Haskell Haskell
      -- | ...
         | VHaskell VHaskell
  deriving (Eq,Data,Typeable)


--
-- basic functions for expanded Haskell
--

-- injecting xhaskell into vhaskell
haskell :: Haskell -> VHaskell
haskell = Obj


--
-- syntactic sugar
-- 

-- building and identifying binary operations
op :: Name -> Haskell -> Haskell -> Haskell
op f l r = App (App (Var ("(" ++ f ++ ")")) l) r

(.+) = op "+"
(.*) = op "*"

isOp :: Name -> Bool
isOp n = head n == '(' && last n == ')'

getOp :: Name -> Name
getOp ('(':o) = take (length o-1) o

-- empty scope dummy value
withoutScope :: Haskell
withoutScope = Var ""

-- variable names
x,y,z :: Haskell
[x,y,z] = map Var ["x","y","z"]

-- choices
choice :: Dim -> [Haskell] -> Haskell
choice d = VHaskell . chc' d 

-- function definitions
fun :: Name -> [Haskell] -> Haskell -> VHaskell
fun n vs e = haskell $ Fun n vs e withoutScope


--
-- adding dimensions
-- 

addPar :: Haskell -> Haskell
addPar (Var "x") = choice "Par" [x,y]
addPar e = e

varyPar :: VHaskell -> VHaskell
varyPar = Dim "Par" ["x","y"] . everywhere (mkT addPar)

addImpl :: Haskell -> Haskell
addImpl e@(App (App (Var "(+)") l) r) | l == r = choice "Impl" [e, Val 2 .* r]
addImpl e = e

varyImpl :: VHaskell -> VHaskell
varyImpl = Dim "Impl" ["plus","times"] . everywhere (mkT addImpl)



--------------
-- Examples --
--------------


-- Twice example.
--
-- twice = Dim "Par" ["x","y"]
--       $ Dim "Impl" ["plus","times"]
--       $ fun "twice" [v] i
--       where v = choice "Par" [x,y]
--             i = choice "Impl" [v .+ v, Val 2 .* v]

-- twice, built directly
twice :: VHaskell
twice = Dim "Par" ["x","y"] $
        Dim "Impl" ["plus","times"] $
        let v = choice "Par" [x,y] in 
            fun "twice" [v] (choice "Impl" [v .+ v, Val 2 .* v])

-- plain twice variants, directly defined
xp, yp, xt, yt :: VHaskell
xp = fun "twice" [x] (x .+ x)
yp = fun "twice" [y] (y .* y)
xt = fun "twice" [x] (Val 2 .* x)
yt = fun "twice" [y] (Val 2 .* y)

-- twice, built from plain expression using transformations
twice'  = (varyPar . varyImpl) xp
twice'' = (varyImpl . varyPar) xp

-- extending Par dimension with another options
twiceZ :: VHaskell
twiceZ = extend "Par" "z" (addAlt (haskell z)) twice


stopTest = extend "Par" "z" (addAlt (haskell z)) twice'
  where dimPar = Dim "Par" ["x","y"]
        twice' = dimPar $
                 Dim "Impl" ["plus","times"] $
                 let v = choice "Par" [x,y] in 
                 let v' = VHaskell $ dimPar (haskell v) in
                 fun "twice" [v] (choice "Impl" [v .+ v', Val 2 .* v])



---------------------
-- Pretty Printing --
---------------------

-- This is not quite perfect because let expressions nested in CC expressions
-- will be rendered as top-level expressions even if they should not be.
-- Good enough for now though.

instance Show Haskell where
  show e = showTop e

showDef n as b = n ++ " " ++ args as ++ " = " ++ showHaskell b
  where args = concat . intersperse " " . map showTerm

showTop (Fun n as b (Var "")) = showDef n as b
showTop (Fun n as b c)        = showDef n as b ++ "\n" ++ showTop c
showTop e                     = showHaskell e

showHaskell (App (App (Var o) l) r) | isOp o = showHaskell  l ++ getOp o ++ showTerm r
showHaskell (App l@(App _ _) r) = showHaskell  l ++ " " ++ showTerm r
showHaskell (App l r)           = showTerm l ++ " " ++ showTerm r
showHaskell (Fun n as b c)      = "let " ++ showDef n as b ++ "in " ++ showHaskell c
showHaskell (VHaskell v)        = show v
showHaskell t                   = showTerm t

showTerm (Var v) = v
showTerm (Val i) = show i
showTerm e@(VHaskell (Chc _ _)) = showHaskell e
showTerm e = "(" ++ showHaskell e ++ ")"

