{-# LANGUAGE DeriveDataTypeable #-}

module Examples.List where

import Data.Generics (Data,Typeable)
import qualified Data.Foldable as F (foldMap,fold)

import CC.Syntax
import CC.Pretty
import CC.Semantics

import Examples.Names
import Examples.Edit


-----------------------
-- Variational Lists --
-----------------------

-- variational lists (vlists)
--
type VList a = V (List a)

-- expanded lists (xlists)
--
data List a = Cons a (List a)
            | Empty
            | VList (VList a) 
            deriving (Eq,Data,Typeable)


--
-- basic functions for expanded lists
--

-- smart constructor for singleton lists
single :: a -> List a
single a = Cons a Empty

-- from a Haskell list to an xlist value
many :: [a] -> List a
many = foldr Cons Empty

-- injecting xlists/ints into vlists/vints
list :: List a -> VList a
list = Obj

int :: Int -> V Int
int = Obj

-- from an xlist value to a Haskell list
fromList :: List a -> [a]
fromList Empty      = []
fromList (Cons a l) = a : fromList l
fromList (VList vl) = fromList' vl

fromList' (Obj l)      = fromList l
fromList' (Dim _ _ vl) = fromList' vl
fromList' (Chc _ vls)  = concatMap fromList' vls

-- lift Haskell list functions to work with xlists
liftL :: ([a] -> b) -> (List a -> b)
liftL = (. fromList)

-- testing for empty list
isNull :: List a -> Bool
isNull = null . fromList


--
-- basic functions for variational lists
--

-- an empty vlist
vempty :: VList a
vempty  = list Empty

-- two smart constructors and functions for consing to a vlist directly
vcons :: a -> VList a -> VList a
vcons x = list . Cons x . VList

vvcons :: V a -> VList a -> VList a
vvcons vx vl = do {x <- vx; vcons x vl}

-- creating a singleton vlist
vsingle :: a -> VList a
vsingle = list . single

-- lifting a Haskell list into a vlist
vlist :: [a] -> VList a
vlist = list . many

-- dimension for optional values
opt :: Dim -> a -> VList a
opt d x = atomic d ["yes","no"] [vsingle x,vempty]


--
-- variational list functions
--

-- foldr
fold :: (a -> b -> b) -> b -> List a -> V b
fold _ b Empty      = obj b
fold f b (Cons a l) = fmap (f a) (fold f b l)
fold f b (VList vl) = vl >>= fold f b
-- fold f b (VList vl) = liftV (fold f b) vl

vfold :: (a -> b -> b) -> b -> VList a -> V b
vfold f b = liftV (fold f b)

-- length
len :: List a -> V Int
len = fold (\_ s->succ s) 0
{-
len :: List a -> V Int
len Empty       = int 0
len (Cons _ xs) = fmap (+1) (len xs)
len (VList vl)  = vl >>= len
-- len (VList vl)  = liftV len vl
-}
vlen :: VList a -> V Int
vlen = liftV len


-- sum
sumL :: List Int -> V Int
sumL = fold (+) 0
{-
sumL :: List Int -> V Int
sumL Empty       = list 0
sumL (Cons x xs) = fmap (x+) (sumL xs)
sumL (VList vl)  = vl >>= sumL
-}
vsum :: VList Int -> V Int
vsum = liftV sumL


-- nth
nth :: Int -> List a -> V a
nth _ Empty       = undefined
nth 1 (Cons x _)  = obj x
nth n (Cons _ xs) = nth (n-1) xs
nth n (VList vl)  = vl >>= nth n

vnth ::  Int -> VList a -> V a
vnth n = liftV (nth n)


-- filter
filterL :: (a -> Bool) -> List a -> List a
filterL p Empty       = Empty
filterL p (Cons x xs) | p x       = Cons x (filterL p xs)
                      | otherwise = filterL p xs
filterL p (VList vl)  = VList (fmap (filterL p) vl)

vfilter :: (a -> Bool) -> VList a -> VList a
vfilter p = fmap (filterL p)


-- map
mapL :: (a -> b) -> List a -> List b
mapL = fmap

vmap :: (a -> b) -> VList a -> VList b
vmap = fmap . fmap
-- vmap f = fmap (mapL f)


-- elem
member :: Eq a => a -> List a -> V Bool
member _ Empty                   = obj False
member x (Cons y ys) | x==y      = obj True
                     | otherwise = member x ys
member x (VList vl)              = vl >>= member x

vmember :: Eq a => a -> VList a -> V Bool
vmember x = liftV (member x)


-- nub
rdup :: Eq a => List a -> List a
rdup Empty       = Empty
rdup (Cons x xs) = Cons x (filterL (/=x) (rdup xs))
rdup (VList vl)  = VList (fmap rdup vl)

vrdup :: Eq a => VList a -> VList a
vrdup = fmap rdup


-- append
cat :: List a -> List a -> List a
cat Empty      r = r
cat (Cons x l) r = Cons x (l `cat` r)
cat (VList vl) r = VList (fmap (`cat` r) vl)

vcat :: VList a -> VList a -> VList a
vcat l r = list $ cat (VList l) (VList r)


-- reverse
rev :: Eq a => List a -> List a
rev Empty       = Empty
rev (Cons x xs) = rev xs `cat` single x 
rev (VList vl)  = VList (fmap rev vl)

vrev :: Eq a => VList a -> VList a
vrev = fmap rev


-- zip
zipL :: List a -> List b -> List (a,b)
zipL Empty       ys          = Empty
zipL xs          Empty       = Empty 
zipL (Cons x xs) (Cons y ys) = Cons (x,y) (zipL xs ys)
zipL (VList vl)  ys          = VList (fmap (`zipL` ys) vl)
zipL xs          (VList vl') = VList (fmap (xs `zipL`) vl')

vzip :: VList a -> VList b -> VList (a,b)
vzip vl vl' = list $ zipL (VList vl) (VList vl')



--------------
-- Examples --
--------------

-- menu example
--
data Food = Steak | Pasta | Fries | Cake
          | Sherry | Broccoli
            deriving (Eq,Show,Data,Typeable)
    
type Menu = VList Food

dessert :: Menu
dessert = opt "Dessert" Cake

menu :: Menu
menu = atomic "Main" ["meat","pasta"] 
              [vlist [Steak,Fries],Pasta `vcons` dessert]

aperitif :: Menu
aperitif = opt "Drink" Sherry

veggies :: Menu
veggies = opt "Vegetables" Broccoli


-- alternative definition
meat  = "meat"  <: vlist [Steak,Fries]
pasta = "pasta" <: Pasta `vcons` dessert
menu2 = alt "Main" [meat,pasta]

-- hoisted Dessert
dMenu = hoist "Dessert" menu

-- poisted and prioritized Dessert
dMenu' = prioritize "Dessert" "Main" dMenu


-- prioritization 
--
ba  = dimB $ dimA $ chcA [vsingle "a","a'" `vcons` chcB [vsingle "b",vsingle "b'"]]
ba' = dimB $ chcB [vlist ["a'","b"],atomicA [vsingle "a",vlist ["a'","b'"]]]


-- some generic examples
--
vx  = chc'A [1,2]
vy  = chc'A [3,4]
vyB = chc'B [3,4]
vxs = vlist [8,9]

vl1 = dimA $ vx `vvcons` vxs
vl2 = dimA $ vx `vvcons` (vy `vvcons` vxs)
vl3 = dimA $ dimB $ vx `vvcons` (vyB `vvcons` vxs)
vl4 = dimA $ vx `vvcons` (dimB $ vyB `vvcons` vxs)



---------------
-- Instances --
---------------

instance Functor List where
  fmap _ Empty       = Empty
  fmap f (Cons x xs) = Cons (f x) (fmap f xs)
  fmap f (VList vl)  = VList $ fmap (fmap f) vl

instance Monad List where
  return = single
  Empty    >>= _ = Empty
  VList e  >>= f = VList (fmap (>>= f) e)
  Cons a l >>= f = let l' = l >>= f in
                   case f a of
                      Empty    -> l'
                      Cons b m -> Cons b (m `cat` l')
                      VList e  -> VList $ fmap (`cat` l') e

instance Show a => Show (List a) where
  show (VList e)  = show e
  show Empty      = "[]"
  show vl         = '[':show' vl++"]"

show' :: Show a => List a -> String
show' (Cons x vl) | isNull vl = show x
                  | otherwise = show x++';':show' vl
show' (VList (Obj l)) = show' l
show' (VList e)       = show e

-- isEmpty Empty = True
-- isEmpty (VList 
-- instance Show a => Show (List a) where
--   show Empty      = "[]"
--   show (Cons a l) = show a ++ ":" ++ show l
--   show (VList e)  = show e

